#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <glib.h>
#include <strings.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#include <stdio.h>
#include <errno.h>

#include "gvfs.h"
#include "dialogs.h"
#include "addons.h"
#include "common.h"
#include "pixmaps.h"
#include "rq.h"
#include "config.h"

static gint return_value = 0;
static gboolean done = FALSE;
typedef enum
  {
    BTN_OK = 1,
    BTN_CANCEL = 2,
    BTN_RETRY = 4,
    BTN_OVERWRITE = 16,
    BTN_OVERWRITE_ALL = 32,
    BTN_IGNORE = 64,
    BTN_IGNORE_ALL = 128,
    BTN_SKIP = 256,
    BTN_SKIP_ALL = 512,
    BTN_APPEND = 1024,
    BTN_APPEND_ALL = 2048
  }
DlgButtons;

gchar *dialog_titles[] =
{
  "GFB message",
  "GFB question",
  "GFB warning",
  "GFB FATAL",
  "New folder creation",
  "Symlink creation",
  "GFB: Copying file(s)",
  "GFB: Deleting file(s)"
};

gchar *std_button_titles[] =
{
  "Ok",
  "Cancel",
  "Retry",
  "Overwrite",
  "Overwrite all",
  "Ignore",
  "Ignore all",
  "Skip",
  "Skip all",
  "Append",
  "Append all"
};

gchar *copy_dlg_strings[] =
{
  "Reading directory",
  "Creating directory",
  "Copying file",
  "From:",
  "To:",
  "Files : %v",
  "Dirs : %v",
  "Bytes: %v",
  " .Done %d files, %d dirs, %d bytes"
};

enum
  {
    CREATE_DIR,
    READ_DIR,
    COPY_FILE,
    FROM,
    TO,
    FILES,
    DIRS,
    BYTES,
    DONE
  };

typedef struct
  {
    gchar *source, *target, *hgc1, *hgc2, *hgc3;
    GList *dirs, *files, *selection, *hGL1, *hGL2;
    GtkWidget *dialog, *from, *status, *file_pb;
    guint sb_id, msg_id, stage;
    gulong files_counter, dirs_counter, bytes_counter;
    gboolean overwrite_all, skip_all, append_all;
  }
gfb_copy_params;

void
ok_button_handler (GtkWidget * widget, gint * data)
{
  done = TRUE;
  *data = DLG_OK;
}

void
cancel_button_handler (GtkWidget * widget, gint * data)
{
  done = TRUE;
  *data = DLG_CANCEL;
}

void
retry_button_handler (GtkWidget * widget, gint * data)
{
  done = TRUE;
  *data = DLG_RETRY;
}

void
ignore_button_handler (GtkWidget * widget, gint * data)
{
  done = TRUE;
  *data = DLG_IGNORE;
}

void
ignore_all_button_handler (GtkWidget * widget, gint * data)
{
  done = TRUE;
  *data = DLG_IGNORE_ALL;
}

void
overwrite_button_handler (GtkWidget * widget, gint * data)
{
  done = TRUE;
  *data = DLG_OVERWRITE;
}

void
overwrite_all_button_handler (GtkWidget * widget, gint * data)
{
  done = TRUE;
  *data = DLG_OVERWRITE_ALL;
}

void
skip_button_handler (GtkWidget * widget, gint * data)
{
  done = TRUE;
  *data = DLG_SKIP;
}

void
skip_all_button_handler (GtkWidget * widget, gint * data)
{
  done = TRUE;
  *data = DLG_SKIP_ALL;
}

void
append_button_handler (GtkWidget * widget, gint * data)
{
  done = TRUE;
  *data = DLG_APPEND;
}

void
append_all_button_handler (GtkWidget * widget, gint * data)
{
  done = TRUE;
  *data = DLG_APPEND_ALL;
}

void
destroy_handler (GtkWidget * widget, gint * data)
{
  done = TRUE;
  *data = DLG_DESTROY;
}

GtkWidget *
ok_button ()
{
  GtkWidget *b = gtk_button_new_with_label (std_button_titles[0]);
  gtk_signal_connect (GTK_OBJECT (b), "clicked",
		      GTK_SIGNAL_FUNC (ok_button_handler), &return_value);
  return b;
}

GtkWidget *
cancel_button ()
{
  GtkWidget *b = gtk_button_new_with_label (std_button_titles[1]);
  gtk_signal_connect (GTK_OBJECT (b), "clicked",
		    GTK_SIGNAL_FUNC (cancel_button_handler), &return_value);
  return b;

}

GtkWidget *
retry_button ()
{
  GtkWidget *b = gtk_button_new_with_label (std_button_titles[2]);
  gtk_signal_connect (GTK_OBJECT (b), "clicked",
		      GTK_SIGNAL_FUNC (retry_button_handler), &return_value);
  return b;
}

GtkWidget *
ignore_button ()
{
  GtkWidget *b = gtk_button_new_with_label (std_button_titles[5]);
  gtk_signal_connect (GTK_OBJECT (b), "clicked",
		    GTK_SIGNAL_FUNC (ignore_button_handler), &return_value);
  return b;
}

GtkWidget *
ignore_all_button ()
{
  GtkWidget *b = gtk_button_new_with_label (std_button_titles[6]);
  gtk_signal_connect (GTK_OBJECT (b), "clicked",
		GTK_SIGNAL_FUNC (ignore_all_button_handler), &return_value);
  return b;
}

GtkWidget *
overwrite_button ()
{
  GtkWidget *b = gtk_button_new_with_label (std_button_titles[3]);
  gtk_signal_connect (GTK_OBJECT (b), "clicked",
		 GTK_SIGNAL_FUNC (overwrite_button_handler), &return_value);
  return b;
}

GtkWidget *
overwrite_all_button ()
{
  GtkWidget *b = gtk_button_new_with_label (std_button_titles[4]);
  gtk_signal_connect (GTK_OBJECT (b), "clicked",
	     GTK_SIGNAL_FUNC (overwrite_all_button_handler), &return_value);
  return b;
}

GtkWidget *
skip_button ()
{
  GtkWidget *b = gtk_button_new_with_label (std_button_titles[7]);
  gtk_signal_connect (GTK_OBJECT (b), "clicked",
		      GTK_SIGNAL_FUNC (skip_button_handler), &return_value);
  return b;
}

GtkWidget *
skip_all_button ()
{
  GtkWidget *b = gtk_button_new_with_label (std_button_titles[8]);
  gtk_signal_connect (GTK_OBJECT (b), "clicked",
		  GTK_SIGNAL_FUNC (skip_all_button_handler), &return_value);
  return b;
}

GtkWidget *
append_button ()
{
  GtkWidget *b = gtk_button_new_with_label (std_button_titles[9]);
  gtk_signal_connect (GTK_OBJECT (b), "clicked",
		    GTK_SIGNAL_FUNC (append_button_handler), &return_value);
  return b;
}

GtkWidget *
append_all_button ()
{
  GtkWidget *b = gtk_button_new_with_label (std_button_titles[10]);
  gtk_signal_connect (GTK_OBJECT (b), "clicked",
		GTK_SIGNAL_FUNC (append_all_button_handler), &return_value);
  return b;
}

GtkWidget *
gfb_dialog_template (DlgButtons buttons)
{
  GtkWidget *dialog;
  done = FALSE;
  return_value = 0;
  dialog = gtk_dialog_new ();
  if ((buttons & BTN_OK) != 0)
    {
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), ok_button (), TRUE, TRUE, 0);
    };
  if ((buttons & BTN_RETRY) != 0)
    {
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), retry_button (), TRUE, TRUE, 0);
    };

  if ((buttons & BTN_OVERWRITE) != 0)
    {
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), overwrite_button (), TRUE, TRUE, 0);
    };

  if ((buttons & BTN_OVERWRITE_ALL) != 0)
    {
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), overwrite_all_button (), TRUE, TRUE, 0);
    };

  if ((buttons & BTN_IGNORE) != 0)
    {
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), ignore_button (), TRUE, TRUE, 0);
    };

  if ((buttons & BTN_IGNORE_ALL) != 0)
    {
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), ignore_all_button (), TRUE, TRUE, 0);
    };

  if ((buttons & BTN_SKIP) != 0)
    {
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), skip_button (), TRUE, TRUE, 0);
    };

  if ((buttons & BTN_SKIP_ALL) != 0)
    {
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), skip_all_button (), TRUE, TRUE, 0);
    };

  if ((buttons & BTN_APPEND) != 0)
    {
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), append_button (), TRUE, TRUE, 0);
    };

  if ((buttons & BTN_APPEND_ALL) != 0)
    {
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), append_all_button (), TRUE, TRUE, 0);
    };

  if ((buttons & BTN_CANCEL) != 0)
    {
      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), cancel_button (), TRUE, TRUE, 0);
    };
  gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
		      GTK_SIGNAL_FUNC (destroy_handler), &return_value);
  gtk_container_set_border_width (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), 10);
  return dialog;
}

gint
gfb_std_dialog (const gchar * message, gint type)
{
  GtkWidget *std_dialog, *pixmap, *label;
  GfbTypeDef *pixmap_index;
  GfbItemIcons *dialog_icon;
  switch (type)
    {
    case DLG_INFORMATION:
      pixmap_index = find_typedef ("GFB_INFO_PIXMAP");
      break;
    case DLG_QUESTION:
      pixmap_index = find_typedef ("GFB_QUESTION_PIXMAP");
      break;
    case DLG_WARNING:
      pixmap_index = find_typedef ("GFB_WARNING_PIXMAP");
      break;
    case DLG_STOP:
      pixmap_index = find_typedef ("GFB_STOP_PIXMAP");
      break;
    default:
      pixmap_index = find_typedef ("GFB_INFO_PIXMAP");
      break;
    };
  dialog_icon = (GfbItemIcons *) pixmap_index->icons;
  std_dialog = gfb_dialog_template (BTN_OK | BTN_CANCEL);
  gtk_grab_add (std_dialog);
  gtk_window_set_title (GTK_WINDOW (std_dialog), dialog_titles[type]);
  pixmap = gtk_pixmap_new (dialog_icon->Icon1, dialog_icon->Bitmap1);
  label = gtk_label_new (message);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (std_dialog)->vbox), pixmap, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (std_dialog)->vbox), label, TRUE, TRUE, 0);
  gtk_widget_show_all (std_dialog);
  while (!done)
    gtk_main_iteration_do (TRUE);
  gtk_grab_remove (std_dialog);
  if (return_value != DLG_DESTROY)
    gtk_widget_destroy (std_dialog);
  return return_value;
}

gchar *
gfb_new_dir ()
{
  GtkWidget *dialog, *input_string;
  GtkWidget *pixmap;
  gchar *result_string;
  dialog = gfb_dialog_template (BTN_OK | BTN_CANCEL);
  gtk_window_set_title (GTK_WINDOW (dialog), dialog_titles[DLG_NEW_FOLDER]);
  input_string = gtk_entry_new_with_max_length (NAME_MAX);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), input_string, TRUE, TRUE, 0);
  gtk_signal_connect (GTK_OBJECT (input_string), "activate", GTK_SIGNAL_FUNC (ok_button_handler), &return_value);
  gtk_widget_set_usize (dialog, 250, 0);
  gtk_widget_show_all (dialog);
  gtk_grab_add (dialog);
  gtk_widget_grab_focus (input_string);
  while (!done)
    gtk_main_iteration_do (TRUE);
  if (return_value == DLG_OK)
    result_string = g_strdup (gtk_entry_get_text (GTK_ENTRY (input_string)));
  else
    result_string = NULL;
  if (return_value != DLG_DESTROY)
    {
      gtk_grab_remove (dialog);
      gtk_widget_destroy (dialog);
    };
  done = FALSE;
  if (result_string != NULL)
    {
      return_value = gvfs_mkdir (result_string);
      if (return_value == -1)
	gfb_std_dialog ("Internal error.\n. Please let authors know", DLG_WARNING);
      else if (return_value != 0)
	gfb_std_dialog (sys_errlist[return_value], DLG_WARNING);
      else
	putRequest (REFRESH_PANELS, NULL, NULL);
      g_free (result_string);
    };
}

gchar *
gfb_new_symlink (gchar * link_target)
{
  GtkWidget *dialog, *link_name, *target_name;
  gchar *result_string;
  dialog = gfb_dialog_template (BTN_OK | BTN_CANCEL);
  gtk_window_set_title (GTK_WINDOW (dialog), dialog_titles[DLG_NEW_SYMLINK]);
  link_name = gtk_entry_new_with_max_length (256);
  target_name = gtk_entry_new_with_max_length (4096);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), link_name, TRUE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), target_name, TRUE, TRUE, 0);
  gtk_signal_connect (GTK_OBJECT (link_name), "activate", GTK_SIGNAL_FUNC (ok_button_handler), &return_value);
  gtk_signal_connect (GTK_OBJECT (target_name), "activate", GTK_SIGNAL_FUNC (ok_button_handler), &return_value);
  if (link_target != NULL)
    {
      gtk_entry_set_text (GTK_ENTRY (target_name), link_target);
      gtk_entry_set_position (GTK_ENTRY (target_name), 0);
      gtk_entry_select_region (GTK_ENTRY (target_name), 0, strlen (link_target));
    }
  else
    gtk_entry_set_text (GTK_ENTRY (target_name), "");
  gtk_widget_set_usize (dialog, 300, 0);
  gtk_widget_show_all (dialog);
  gtk_grab_add (dialog);
  gtk_widget_grab_focus (link_name);
  while (!done)
    gtk_main_iteration_do (TRUE);
  if (return_value == DLG_OK)
    {
      result_string = g_strdup (gtk_entry_get_text (GTK_ENTRY (link_name)));
      link_target = g_strdup (gtk_entry_get_text (GTK_ENTRY (target_name)));
    }
  else
    result_string = NULL;
  done = FALSE;
  if (return_value != DLG_DESTROY)
    {
      gtk_grab_remove (dialog);
      gtk_widget_destroy (dialog);
    };
  if (result_string != NULL)
    {
      return_value = gvfs_symlink (result_string, link_target);
      if (return_value == -1)
	gfb_std_dialog ("Internal error.\n Please let authors know", DLG_WARNING);
      else if (return_value != 0)
	gfb_std_dialog (sys_errlist[return_value], DLG_WARNING);
      else
	putRequest (REFRESH_PANELS, NULL, NULL);
      g_free (result_string);
      g_free (link_target);
    };
}

void
prepare_fd_list (GList * selection, GList ** dirs, GList ** files)
{
  UniversalItem *hUI1 = NULL;
  UniversalItem *hUI2 = NULL;
  FileInfo *hFI1 = NULL;
  GList *hGL1 = selection;
//initial list buildup
  while (hGL1 != NULL)
    {
      hUI1 = UNI_ITEM (hGL1->data);
      switch (hUI1->type)
	{
	case FILE_LIST_DIR:
	case FILE_LIST_FILESYSTEM:
	  hUI2 = copy_universalitem (hUI1);
	  hUI2->data = copy_fileinfo (hUI1->data);
	  *dirs = g_list_append (*dirs, hUI2);
	  break;
	case FILE_LIST_FILE:
	  hUI2 = copy_universalitem (hUI1);
	  hUI2->data = copy_fileinfo (hUI1->data);
	  *files = g_list_append (*files, hUI2);
	  break;
	case TREE_NODE_DIR:
	case TREE_NODE_FILESYSTEM:
	  hUI2 = g_malloc0 (sizeof (UniversalItem));
	  hUI2->type = FILE_LIST_DIR;
	  hFI1 = g_malloc0 (sizeof (FileInfo));
	  hFI1->name = g_strdup ((gchar *) (hUI1->data));
	  hFI1->selected = TRUE;
	  hUI2->data = hFI1;
	  *dirs = g_list_append (*dirs, hUI2);
	  break;
	};
      hGL1 = hGL1->next;
    };
}

GtkWidget *
gfb_copy_dialog (GtkWidget * sb)
{
  GtkWidget *hbox, *vbox;
  GtkWidget *dialog = gtk_dialog_new ();
  vbox = gtk_vbox_new (FALSE, 0);
  hbox = gtk_hbox_new (TRUE, 5);
  gtk_container_remove (GTK_CONTAINER (dialog), GTK_DIALOG (dialog)->vbox);
  GTK_DIALOG (dialog)->vbox = vbox;
  GTK_DIALOG (dialog)->action_area = hbox;
  gtk_container_add (GTK_CONTAINER (dialog), vbox);
  gtk_box_pack_end (GTK_BOX (vbox), sb, FALSE, FALSE, 0);
  gtk_box_pack_end (GTK_BOX (vbox), hbox, FALSE, TRUE, 0);
  gtk_box_pack_end (GTK_BOX (vbox), gtk_hseparator_new (), FALSE, TRUE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), cancel_button (), FALSE, FALSE, 0);
  gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
		      GTK_SIGNAL_FUNC (destroy_handler), &return_value);
  return dialog;
}

void
zero_copy_params (gfb_copy_params * p)
{
  p->source = p->target = p->hgc1 = p->hgc2 = p->hgc3 = NULL;
  p->dirs = p->files = p->selection = p->hGL1 = p->hGL2 = NULL;
  p->dialog = p->status = p->file_pb = p->from = 0;
  p->sb_id = p->msg_id = p->stage = 0;
  p->files_counter = p->bytes_counter = p->dirs_counter = 0;
  p->overwrite_all = FALSE;
  p->skip_all = FALSE;
  p->append_all = FALSE;
  p->hgc2 = g_malloc0 (256);	// Just a number. No special meaning

}

void
create_copy_dialog (gfb_copy_params * p)
{
  p->from = gtk_label_new ("");
  gtk_label_set_justify (GTK_LABEL (p->from), GTK_JUSTIFY_LEFT);
  p->status = gtk_statusbar_new ();
  p->sb_id = gtk_statusbar_get_context_id (GTK_STATUSBAR (p->status), "Copying");
  p->file_pb = gtk_progress_bar_new ();
  gtk_progress_bar_set_bar_style (GTK_PROGRESS_BAR (p->file_pb), GTK_PROGRESS_DISCRETE);
  gtk_progress_bar_set_discrete_blocks (GTK_PROGRESS_BAR (p->file_pb), 50);
  gtk_misc_set_alignment (GTK_MISC (p->from), 0.0, 0.1);
  gtk_widget_set_usize (p->file_pb, 0, 8);
// Creating dialog window
  p->dialog = gfb_copy_dialog (p->status);
  gtk_window_set_title (GTK_WINDOW (p->dialog), dialog_titles[DLG_COPYING]);
//Packing widgets
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG (p->dialog)->vbox), p->from, FALSE, FALSE, 0);
  gtk_box_pack_end (GTK_BOX (GTK_DIALOG (p->dialog)->vbox), p->file_pb, FALSE, FALSE, 0);
  gtk_widget_set_usize (GTK_DIALOG (p->dialog)->vbox, 0, 80);
  gtk_widget_set_usize (p->dialog, 320, 0);
  gtk_widget_show_all (p->dialog);
}

void
set_copy_status (gfb_copy_params * p)
{
  gchar *hgc3 = g_malloc0 (strlen (copy_dlg_strings[p->stage]) + strlen (p->hgc2) + 1);
  strcpy (hgc3, copy_dlg_strings[p->stage]);
  strcat (hgc3, p->hgc2);
  if (p->msg_id > 0)
    gtk_statusbar_remove (GTK_STATUSBAR (p->status), p->sb_id, p->msg_id);
  p->msg_id = gtk_statusbar_push (GTK_STATUSBAR (p->status), p->sb_id, hgc3);
  g_free (hgc3);
}

void
update_copy_progress (gfb_copy_params * p)
{
  snprintf (p->hgc2, 255, copy_dlg_strings[DONE], p->files_counter, p->dirs_counter, p->bytes_counter);
  set_copy_status (p);
}

void
gfb_copy_file (gfb_copy_params * p)
{
  FileInfo *hFI1;
  gvfs_file_copy_struct gfcs;
  GtkWidget *dialog;
  gchar *hgc1;
  gint hgi1;
  gint t_return_value;		/* Need to do stack-like saving of intermediate results */
  gboolean t_done;		/* As for prev var */
  hFI1 = UNI_ITEM (p->files->data)->data;
  if (hFI1->info.st_nlink & 0100000)
    {
      p->hgc1 = g_malloc0 (PATH_MAX);
      gvfs_readlink (hFI1->name, p->hgc1, PATH_MAX - 1);
      p->hgc3 = replace_pattern (hFI1->name, p->source, p->target);
      gvfs_symlink (p->hgc3, p->hgc1);
      g_free (p->hgc1);
      g_free (p->hgc3);
      hFI1->info.st_size = 0;
      return;
    };

  gfcs.source_name = hFI1->name;
  gfcs.target_name = replace_pattern (hFI1->name, p->source, p->target);
  gfcs.total_bytes = hFI1->info.st_size;
  gfcs.done_bytes = 0;
  gfcs.source_gvfs = gfcs.target_gvfs = NULL;
  gfcs.buffer = gfcs.source_fd = gfcs.target_fd = NULL;
  gfcs.force_overwrite = p->overwrite_all;
  gfcs.all_ok = TRUE;
  gfcs.mode = hFI1->info.st_mode;
  while (gfcs.done_bytes < gfcs.total_bytes)
    {
      hgi1 = gvfs_copy_file (&gfcs);
      if (hgi1 != 0)
	{
	  if ((hgi1 == EEXIST) && (p->skip_all == TRUE))
	    gfcs.done_bytes = gfcs.total_bytes;
	  else
	    {
	      t_return_value = return_value;
	      t_done = done;
	      hgc1 = g_malloc (4096);
	      sprintf (hgc1, "There's a problem: \n %s", sys_errlist[hgi1]);
	      dialog = gfb_dialog_template (
					     BTN_RETRY | BTN_OVERWRITE | BTN_OVERWRITE_ALL | BTN_SKIP | BTN_SKIP_ALL | BTN_CANCEL);
	      gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), gtk_label_new (hgc1), TRUE, TRUE, 0);
	      gtk_grab_add (dialog);
	      gtk_widget_show_all (dialog);
	      while (!done)
		{
		  while (gtk_events_pending ())
		    gtk_main_iteration ();
		};
	      switch (return_value)
		{
		case DLG_OVERWRITE:
		  gfcs.force_overwrite = TRUE;
		  gfcs.all_ok = TRUE;
		  break;
		case DLG_OVERWRITE_ALL:
		  gfcs.force_overwrite = TRUE;
		  p->overwrite_all = TRUE;
		  gfcs.all_ok = TRUE;
		  break;
		case DLG_RETRY:
		  gfcs.all_ok = TRUE;
		  break;
		case DLG_SKIP:
		  gfcs.done_bytes = gfcs.total_bytes;
		  gfcs.all_ok = FALSE;
		  break;
		case DLG_SKIP_ALL:
		  gfcs.done_bytes = gfcs.total_bytes;
		  p->skip_all = TRUE;
		  gfcs.all_ok = FALSE;
		  break;
		case DLG_CANCEL:
		case DLG_DESTROY:
		  t_done = TRUE;
		  t_return_value = DLG_CANCEL;
		  gfcs.all_ok = FALSE;
		  gfcs.done_bytes = gfcs.total_bytes;
		  break;
		};
	      g_free (hgc1);
	      if (return_value != DLG_DESTROY)
		{
		  gtk_grab_remove (dialog);
		  gtk_widget_destroy (dialog);
		};
	      done = t_done;
	      return_value = t_return_value;
	    };
	}
      else
	{
	  gtk_progress_configure (GTK_PROGRESS (p->file_pb), gfcs.done_bytes, 0, gfcs.total_bytes);
	  while (gtk_events_pending ())
	    gtk_main_iteration ();
	};
    };
  g_free (gfcs.target_name);
}

gint
gfb_copy_wrapper (GList * data_struct, gint operation)
{
  FileInfo *hFI1;
// First version of this procedure included all of the above 6 functions and was really HUGE ,
  // so I've divided it to smaller parts. Now it's relatively small end easy to understand.
  gfb_copy_params p;		//for easy sharing of otherwise global variables;
  //Zeroing local vars

  zero_copy_params (&p);
  done = FALSE;
  return_value = 0;
// Retrieving data      
  p.source = data_struct->data;
  data_struct = data_struct->next;
  p.target = data_struct->data;
  data_struct = data_struct->next;
  p.selection = data_struct->data;
// Building and shaping widgets
  create_copy_dialog (&p);
// Dirty hack
  if (operation == DELETE_FILES)
    gtk_window_set_title (GTK_WINDOW (p.dialog), dialog_titles[DLG_DELETING]);
  gtk_grab_add (p.dialog);
  prepare_fd_list (p.selection, &p.dirs, &p.files);
  p.hGL1 = p.dirs;
  while (!done)
    {
      if (p.files != NULL)
	{
	  p.stage = COPY_FILE;
	  set_copy_status (&p);
	  hFI1 = UNI_ITEM (p.files->data)->data;
	  gtk_label_set_text (GTK_LABEL (p.from), hFI1->name);
//                                      Copy file pointed by p.files->data from p.source to p.target.
	  if (operation != DELETE_FILES)
	    gfb_copy_file (&p);
	  if ((operation == CUT) || (operation == DELETE_FILES))
	    gvfs_rm (hFI1->name);	// Delete file if it was CUT'N'PASTE or DELETE_FILES

	  p.files_counter++;
	  p.bytes_counter += hFI1->info.st_size;
	  update_copy_progress (&p);
	  g_free (p.files->data);
	  p.files = p.files->next;
	}
      else if (p.hGL1 != NULL)
	{
	  p.stage = READ_DIR;
	  set_copy_status (&p);
//                              Read  directory pointed by p.hGL1->data and append it to p.dirs
	  hFI1 = FILE_INFO (UNI_ITEM (p.hGL1->data)->data);
	  if (!(hFI1->info.st_nlink & 0100000))
	    {
	      read_dir (hFI1->name, &p.dirs, &p.files);
	      p.hgc1 = replace_pattern (hFI1->name, p.source, p.target);
	      gvfs_mkdir (p.hgc1);
	      g_free (p.hgc1);
	    }
	  else
	    {
	      p.hgc1 = g_malloc0 (PATH_MAX);
	      gvfs_readlink (hFI1->name, p.hgc1, PATH_MAX - 1);
	      p.hgc3 = replace_pattern (hFI1->name, p.source, p.target);
	      gvfs_symlink (p.hgc3, p.hgc1);
	      g_free (p.hgc1);
	      g_free (p.hgc3);
	    }
	  p.dirs_counter++;
	  update_copy_progress (&p);
	  p.hGL1 = p.hGL1->next;
	}
      else
	{
	  done = TRUE;
	};
      while (gtk_events_pending ())
	gtk_main_iteration ();
    };
// Destroying dialog window
  if (return_value != DLG_DESTROY)
    {
      gtk_grab_remove (p.dialog);
      gtk_widget_destroy (p.dialog);
    };
//freeing lists. We're doing it from bottom to top because of rmdir.
  for (p.dirs = g_list_last (p.dirs); p.dirs != NULL; p.dirs = p.dirs->prev)
    {
      hFI1 = FILE_INFO (UNI_ITEM (p.dirs->data)->data);
      if (hFI1->name != NULL)
	{
	  if ((operation == CUT) || (operation == DELETE_FILES))
	    {
	      if (!(hFI1->info.st_nlink & 0100000))
		gvfs_rmdir (hFI1->name);	//Remove old directory;

	      else
		gvfs_rm (hFI1->name);	//Remove symlink to dir;

	    };
	  g_free (hFI1->name);
	};
      g_free (p.dirs->data);
    };
  g_list_free (p.dirs);
  g_free (p.source);
  g_free (p.target);
  g_free (p.hgc2);
//Return
  putRequest (REFRESH_PANELS, NULL, NULL);
  return return_value;
}
